"""
Created 18th June 2019
"""
import logging

from spine_aws_common.log.logutil import write_log
from spine_aws_common.log.masking import mask_pid


SEVERITY_INPUT_MAP = {
    "CRITICAL": logging.CRITICAL,
    "ERROR": logging.ERROR,
    "WARN": logging.WARN,
    "INFO": logging.INFO,
    # Allow us to enable python debug independently
    "DEBUG": logging.INFO,
    # Allow us to enable python debug independently
    "TRACE": logging.DEBUG,
}


class LoggingAdapter(logging.Handler):
    """
    Adapter to allow libraries that use python logging to output to our logger
    """

    CRITICAL = "UTI9998"
    ERROR = "UTI9997"
    WARN = "UTI9996"
    INFO = "UTI9995"
    DEBUG = "UTI9994"
    TRACE = "UTI9993"

    _EXCLUSIVE_USE_CHANNEL_FINDER = "in exclusive use"
    _EXCLUSIVE_USE_CHANNEL_OUTPUT = "Connection blocked due to exclusive use"

    _EXCLUSIVE_USE_CALLBACK_FINDER = "bound method BlockingChannel._on_channel_closed"
    _EXCLUSIVE_USE_CALLBACK_OUTPUT = "Callback terminated due to exclusive use"

    # Convert tuples to improve message or adjust criticality
    # E.g.
    # 'pika.connection' reports CRITICAL 'Attempted to send frame when closed'
    # - this is not CRITICAL - SPII-13851

    # Use to replace log message when substring found
    _MESSAGE_MAP = {
        _EXCLUSIVE_USE_CHANNEL_FINDER: _EXCLUSIVE_USE_CHANNEL_OUTPUT,
        _EXCLUSIVE_USE_CALLBACK_FINDER: _EXCLUSIVE_USE_CALLBACK_OUTPUT,
    }

    # Use for mapping logs with a specific message
    _LOGGER_DETAIL_MAP = {
        (
            "pika.connection",
            logging.CRITICAL,
            "CRITICAL",
            "Attempted to send frame when closed",
        ): (
            "pika.connection",
            logging.ERROR,
            "ERROR",
            "Attempted to send frame when closed",
        ),
        ("pika.channel", logging.WARNING, "WARNING", _EXCLUSIVE_USE_CHANNEL_OUTPUT): (
            "pika.channel",
            logging.INFO,
            "INFO",
            _EXCLUSIVE_USE_CHANNEL_OUTPUT,
        ),
        ("pika.callback", logging.ERROR, "ERROR", _EXCLUSIVE_USE_CALLBACK_OUTPUT): (
            "pika.callback",
            logging.INFO,
            "INFO",
            _EXCLUSIVE_USE_CALLBACK_OUTPUT,
        ),
        (
            "pika.adapters.base_connection",
            logging.CRITICAL,
            "CRITICAL",
            "Tried to handle an error where no error existed",
        ): (
            "pika.adapters.base_connection",
            logging.WARN,
            "WARN",
            "Tried to handle an error where no error existed",
        ),
    }

    # Use for mapping of all logs of a given name
    _LOGGER_SUMMARY_MAP = {
        ("tornado.access", logging.INFO, "INFO"): (
            "tornado.access",
            logging.DEBUG,
            "DEBUG",
        )
    }

    def __init__(self, log_object=None):
        logging.Handler.__init__(self, logging.DEBUG)
        self.log_object = log_object

    def emit(self, record):
        """Override emit to output to our log file"""
        name, level, levelname, message = self._switch_log(
            record.name, record.levelno, record.levelname, record.getMessage()
        )

        masked_message = mask_pid(message)
        log_reference = self._get_log_reference(level)
        log_dict = {"logger": name, "message": masked_message, "level": levelname}

        # Note that we don't log sys.exc_info() here since there is no
        # guarantee that a third party library is logging in response to an
        # exception. This has lead to misleading stack traces being output
        # that were nothing to do with the logged message.
        if self.log_object:
            self.log_object.write_log(log_reference, None, log_dict)
        else:
            # Needs to be improved.
            if "logReference=" in message:
                return
            write_log(log_reference, None, log_dict)

    def _switch_log(self, name, level, levelname, message):
        """
        If the log is in the dictionary, return the dictionary output instead
        """
        name, level, levelname = LoggingAdapter._LOGGER_SUMMARY_MAP.get(
            (name, level, levelname), (name, level, levelname)
        )

        for finder, value in self._MESSAGE_MAP.items():
            if finder in message:
                message = value

        return LoggingAdapter._LOGGER_DETAIL_MAP.get(
            (name, level, levelname, message), (name, level, levelname, message)
        )

    def _get_log_reference(self, level):
        """
        Get the log reference based on the output map
        """
        output_map = {
            logging.CRITICAL: self.CRITICAL,
            logging.FATAL: self.CRITICAL,
            logging.ERROR: self.ERROR,
            logging.WARNING: self.WARN,
            logging.WARN: self.WARN,
            logging.INFO: self.INFO,
            # Allow us to enable python debug independently
            logging.DEBUG: self.TRACE,
        }

        return output_map[level]


def configure_third_party_logging_adapter(severity_threshold):
    """
    Configure an adapter to allow libraries that use standard Python logging to
    output to our log files.
    """
    root_logger = logging.getLogger()
    if root_logger.handlers:
        for handler in root_logger.handlers:
            root_logger.removeHandler(handler)
    root_logger.setLevel(SEVERITY_INPUT_MAP[severity_threshold])

    adapter = LoggingAdapter()
    root_logger.addHandler(adapter)



"""
New Logging Attempts
"""
from spine_aws_common.log.spinelogging import SpineLogger
from spine_aws_common.log.constants import LoggingConstants
from spine_aws_common.log.details import get_log_details
from spine_aws_common.log.formatting import (
    add_default_keys,
    create_log_preamble,
    evaluate_log_keys,
    substitute_preamble_for_monitor,
)
from spine_aws_common.log.masking import mask_url
from spine_aws_common.log.writer import write_to_file


def write_log(
    log_reference="UTI9999",
    error_list=None,
    log_row_dict=None,
    severity_threshold_override=None,
    process_name=None,
):
    """
    The writing of the log allows the following information to be passed:
    :param log_reference: this should resolve to a logReference within the
    logBase
    :type  log_reference: str
    :param error_list: the output of a sys.exc_info() where an exception has
    been caught
    :param log_row_dict - a dictionary of substitutions to be made against the
    logText in the logReference
    :type log_row_dict: dict
    :param severity_threshold_override: Not normally present - allows the
    standard log level to be over-ridden for this entry
    :param process_name: Not normally present - allows the standard processName
    to be over-ridden for this entry
    The process for writing a log file entry is:
    Lookup the log reference information in the log base
    Exit out if the log level of the log is above that at which the user is set
    to log (e.g. if it is a DEBUG log and the user level is set to INFO)
    Create an audit version of the log_row_dict containing sensitive data, and
    determine if an Audit entry is required.
    """
    if log_row_dict is None:
        log_row_dict = {}

    if process_name is None:
        process_name = SpineLogger.get_process_name()

    log_details = get_log_details(
        log_reference, SpineLogger.get_log_base_dict(), SpineLogger.get_log_base_cache()
    )
    if not log_details or not log_details.check_log_severity_for_log(
        severity_threshold_override, SpineLogger.get_severity_threshold()
    ):
        return None

    # If not provided, set empty values for internalID and sessionId
    add_default_keys(log_row_dict)

    evaluate_log_keys(log_details, log_row_dict)

    log_preamble = create_log_preamble(
        log_details.log_level, process_name, log_reference
    )
    log_row_dict_masked = mask_url(log_row_dict)

    if log_details.audit_log_required:
        write_to_file(
            log_preamble,
            log_details.log_text,
            log_row_dict_masked,
            LoggingConstants.LFR_AUDIT,
        )
    else:
        write_to_file(
            log_preamble,
            log_details.log_text,
            log_row_dict_masked,
            LoggingConstants.LFR_OPERATIONS,
        )

    if log_details.monitor_log_required:
        # Swap to Log_Level=MONITOR - will help prevent SALTing requirement
        # As Splunk may get matching CRC check for Audit and Monitor Log.
        write_to_file(
            substitute_preamble_for_monitor(log_preamble),
            log_details.log_text,
            log_row_dict_masked,
            LoggingConstants.LFR_NMS,
        )

    if log_details.check_log_severity_for_crashdump(
        severity_threshold_override, SpineLogger.get_severity_threshold(), error_list
    ):
        stub_log_reference = LoggingConstants.LR_CRASHDUMP
        stub_log_details = get_log_details(
            stub_log_reference,
            SpineLogger.get_log_base_dict(),
            SpineLogger.get_log_base_cache(),
        )
        stub_log_preamble = create_log_preamble(
            stub_log_details.log_level, process_name, stub_log_reference
        )

        # Write stub crashdump to spinevfmoperations, so that non-SC cleared
        # staff can see a crash occurred.
        write_to_file(
            stub_log_preamble,
            stub_log_details.log_text,
            {"originalLogReference": log_reference},
            LoggingConstants.LFR_OPERATIONS,
        )

        # Write actual crashdump.
        write_to_file(
            log_preamble,
            log_details.log_text,
            log_row_dict,
            LoggingConstants.LFR_CRASHDUMP,
            error_list,
        )

    return log_details.log_text



"""
Base batch Lambda application
"""
from abc import abstractmethod

from spine_aws_common.lambda_application import LambdaApplication


class BatchApplication(LambdaApplication):
    """
    Base class for Batch Lambda applications
    """

    def __init__(self, additional_log_config=None, load_ssm_params=False):
        super().__init__(
            additional_log_config=additional_log_config, load_ssm_params=load_ssm_params
        )
        self.records = None

    def initialise(self):
        """
        Application initialisation
        """
        self.records = getattr(self.event, "records", [])

    def start(self):
        """
        Start the application
        """
        for record in self.records:
            self.log_object.set_internal_id(self._get_internal_id_from_record(record))
            self.process_record(record)

    def _get_internal_id_from_record(self, record):
        """
        Get (or create new) internalID from record
        """
        return record.get("internal_id", self._create_new_internal_id())

    @abstractmethod
    def process_record(self, record):
        """
        Process a single record from the batch
        """



"""Common methods and classes used for mesh client"""
from collections import namedtuple
import os
import json
import boto3
from mesh_client import MeshClient


class SingletonCheckFailure(Exception):
    """Singleton check failed"""

    def __init__(self, msg=None):
        super().__init__()
        self.msg = msg


class AwsFailedToPerformError(Exception):
    """Errors raised by AWS functions"""

    def __init__(self, msg=None):
        super().__init__()
        self.msg = msg


class MeshCommon:
    """Common"""

    MIB = 1024 * 1024
    DEFAULT_CHUNK_SIZE = 20 * MIB

    @staticmethod
    def singleton_check(mailbox, my_step_function_name):
        """Find out whether there is another step function running for my mailbox"""
        sfn_client = boto3.client("stepfunctions", region_name="eu-west-2")
        response = sfn_client.list_state_machines()
        # Get my step function arn
        my_step_function_arn = None
        for step_function in response.get("stateMachines", []):
            if step_function.get("name", "") == my_step_function_name:
                my_step_function_arn = step_function.get("stateMachineArn", None)

        # TODO add this check to tests
        if not my_step_function_arn:
            raise SingletonCheckFailure(
                "No executing step function arn for "
                + f"step_function={my_step_function_name}"
            )

        response = sfn_client.list_executions(
            stateMachineArn=my_step_function_arn,
            statusFilter="RUNNING",
        )
        currently_running_step_funcs = []
        for execution in response["executions"]:
            currently_running_step_funcs.append(execution["executionArn"])

        exec_count = 0
        for execution_arn in currently_running_step_funcs:
            response = sfn_client.describe_execution(executionArn=execution_arn)
            step_function_input = json.loads(response.get("input", "{}"))
            input_mailbox = step_function_input.get("mailbox", None)
            if input_mailbox == mailbox:
                exec_count = exec_count + 1
            if exec_count > 1:
                raise SingletonCheckFailure("Process already running for this mailbox")

        return True

    @staticmethod
    def convert_params_to_dict(params):
        """Convert paramater dict to key:value dict"""
        new_dict = {}
        for entry in params:
            name = entry.get("Name", None)
            if name:
                var_name = os.path.basename(name)
                new_dict[var_name] = entry.get("Value", None)
        return new_dict

    @staticmethod
    def return_failure(log_object, status, logpoint, mailbox, message=""):
        """Return a failure response with retry"""
        log_object.write_log(logpoint, None, {"mailbox": mailbox, "error": message})
        return {
            "statusCode": status,
            "headers": {
                "Content-Type": "application/json",
                "Retry-After": 18000,
            },
            "body": {
                "internal_id": log_object.internal_id,
                "error": message,
            },
        }

    @staticmethod
    def get_ssm_params(path, recursive=False, decryption=True):
        """Get parameters from SSM param store and return as simple dict"""
        # TODO region name fix
        ssm_client = boto3.client("ssm", region_name="eu-west-2")

        params_result = ssm_client.get_parameters_by_path(
            Path=path,
            Recursive=recursive,
            WithDecryption=decryption,
        )
        params = params_result.get("Parameters", {})
        new_params_dict = {}
        for entry in params:
            name = entry.get("Name", None)
            if name:
                var_name = os.path.basename(name)
                new_params_dict[var_name] = entry.get("Value", None)
        return new_params_dict


class ExtendedMeshClient(MeshClient):
    """Extended functionality for lambda send"""


# Named tuple for holding Mesh Message info
MeshMessage = namedtuple(
    "MeshMessage",
    [
        "filename",
        "body",
        "src_mailbox",
        "dest_mailbox",
        "workflow_id",
        "message_id",
    ],
)



"""
Common utilities used in lambda application classes
"""
from datetime import datetime, timedelta


class StopWatch:
    """
    Class to support timing points in the code
    """

    def __init__(self):
        self.start_time = None

    def start_the_clock(self):
        """
        Start the clock
        """
        self.start_time = datetime.now()

    def stop_the_clock(self):
        """
        Stop the clock automatically resets and restarts the clock
        Use split the clock if want to keep a parent timer running
        """
        step_duration_seconds = self.split_the_clock()
        self.start_time = datetime.now()
        return step_duration_seconds

    def split_the_clock(self):
        """
        Split the clock, keeping the parent timer running
        """
        step_duration = datetime.now() - self.start_time
        step_duration_seconds = round(
            float(step_duration.seconds) + float(step_duration.microseconds) / 1000000,
            3,
        )
        if step_duration_seconds < 0.0005:
            step_duration_seconds = 0.000

        return step_duration_seconds

    def reset_the_clock(self, seed_time):
        """
        Reset the clock assuming a new seed time, to be used when time has
        been passed as message by string Assumed format of time is:
        %Y%m%dT%H%M%S.%3N
        """
        date_split = seed_time.split(".")
        self.start_time = datetime.strptime(date_split[0], "%Y%m%dT%H%M%S")
        self.start_time += timedelta(milliseconds=int(date_split[1]))


def human_readable_bytes(num):
    """Size in human readable format for logs"""
    if abs(num) < 1024:
        return f"{num}B"
    num /= 1024.0
    for unit in ["Ki", "Mi", "Gi", "Ti"]:
        if abs(num) < 1024.0:
            return f"{num:.1f}{unit}B"
        num /= 1024.0
    return f"{num:.1f}PiB"



"""
Created 18th June 2019
"""


class LoggingConstants:
    # pylint:disable=too-few-public-methods
    """
    Logging Constants
    """

    SECTION_LEVEL = "Log Level"
    SECTION_TEXT = "Log Text"

    LFR_OPERATIONS = "operations"
    LFR_AUDIT = "audit"
    LFR_CRASHDUMP = "crashdump"
    LFR_CONSOLE = "console"
    LFR_NMS = "monitor"

    LOGGING_DIRECTORIES = [LFR_AUDIT, LFR_OPERATIONS, LFR_CRASHDUMP, LFR_NMS]

    TRACE = 8
    DEBUG = 7
    INFO = 6
    WARN = 4
    ERROR = 3
    CRITICAL = 1
    AUDIT = 0

    LR_CRASHDUMP = "UTI9992"
    LOG_SUFFIX = ".log"

    IDENTIFIERS = ["internalID", "sessionid"]

    SPINE_LOGGER = "spine"
